package prev.phase.seman;

import prev.common.report.*;
import prev.data.ast.tree.*;
import prev.data.ast.tree.decl.*;
import prev.data.ast.tree.expr.*;
import prev.data.ast.tree.stmt.*;
import prev.data.ast.tree.type.*;
import prev.data.ast.visitor.*;

/**
 * Name resolver.
 * 
 * Name resolver connects each node of a abstract syntax tree where a name is
 * used with the node where it is declared. The only exceptions are a record
 * field names which are connected with its declarations by type resolver. The
 * results of the name resolver are stored in
 * {@link prev.phase.seman.SemAn#declaredAt}.S
 */
public class NameResolver extends AstFullVisitor<Object, NameResolver.Mode> {

	public enum Mode {
		HEAD, BODY
	}

	private SymbTable symbTable = new SymbTable();

	@Override
	public Object visit(AstTrees<? extends AstTree> trees, Mode mode) {
		for (AstTree t : trees) {
			if (t != null) {
				t.accept(this, Mode.HEAD);
			}
		}

		for (AstTree t : trees) {
			if (t != null) {
				t.accept(this, Mode.BODY);
			}
		}

		return null;
	}

	@Override
	public Object visit(AstNameType nameType, Mode mode) {
		if(mode == Mode.BODY) {
			try {
				SemAn.declaredAt.put(nameType, symbTable.fnd(nameType.name));
			}
			catch (SymbTable.CannotFndNameException e) {
				throw new Report.Error(nameType, "Semantic error: cannot find name " + nameType.name);
			}
		}

		return null;
	}

	@Override
	public Object visit(AstFunDecl funDecl, NameResolver.Mode mode) {
		if(mode == Mode.HEAD) {
			try {
				symbTable.ins(funDecl.name, funDecl);
			}
			catch (SymbTable.CannotInsNameException e) {
				throw new Report.Error(funDecl, "Semantic error: redeclaration of function " + funDecl.name);
			}
		}
		else if(mode == Mode.BODY) {
			// funDecl.pars.accept(this, mode);
			funDecl.type.accept(this, mode);

			symbTable.newScope();

			if(funDecl.pars != null) {
				funDecl.pars.accept(this, Mode.HEAD);
			}

			if(funDecl.expr != null) {
				funDecl.expr.accept(this, Mode.BODY);
			}

			symbTable.oldScope();
		}

		return null;
	}

	@Override
	public Object visit(AstRecExpr recExpr, Mode mode) {
		if(mode == Mode.BODY) {
			recExpr.rec.accept(this, mode);
		}

		return null;
	}

	@Override
	public Object visit(AstParDecl parDecl, Mode mode) {
		if(mode == Mode.HEAD) {
			try {
				symbTable.ins(parDecl.name, parDecl);
			}
			catch (SymbTable.CannotInsNameException e) {
				throw new Report.Error(parDecl, "Semantic error: redeclaration of function parameter " + parDecl.name);
			}
		}
		else if (mode == Mode.BODY) {
			parDecl.type.accept(this, mode);
		}

		return null;
	}

	@Override
	public Object visit(AstTypeDecl typeDecl, Mode mode) {
		if (mode == Mode.HEAD) {
			try {
				symbTable.ins(typeDecl.name, typeDecl);
			}
			catch (SymbTable.CannotInsNameException e) {
				throw new Report.Error(typeDecl, "Semantic error: redeclaration of type " + typeDecl.name);
			}
		}
		else if (mode == Mode.BODY) {
			typeDecl.type.accept(this, mode);
		}

		return null;
	}

	@Override
	public Object visit(AstVarDecl varDecl, Mode mode) {
		if (mode == Mode.HEAD) {
			try {
				symbTable.ins(varDecl.name, varDecl);
			}
			catch (SymbTable.CannotInsNameException e) {
				throw new Report.Error(varDecl, "Semantic error: redeclaration of variable " + varDecl.name);
			}
		}
		else if (mode == Mode.BODY) {
			varDecl.type.accept(this, mode);
		}

		return null;
	}

	@Override
	public Object visit(AstWhereExpr whereExpr, Mode mode) {
		if (mode == Mode.BODY) {
			symbTable.newScope();

			whereExpr.decls.accept(this, Mode.HEAD);
			whereExpr.expr.accept(this, Mode.BODY);

			symbTable.oldScope();
		}

		return null;
	}

	@Override
	public Object visit(AstCallExpr callExpr, Mode mode) {
		if (mode == Mode.BODY) {
			try {
				SemAn.declaredAt.put(callExpr, symbTable.fnd(callExpr.name));
			}
			catch (SymbTable.CannotFndNameException e) {
				throw new Report.Error(callExpr, "Semantic error: cannot find name " + callExpr.name);
			}

			if(!(callExpr.args == null)) {
				callExpr.args.accept(this, mode);
			}
		}

		return null;
	}

	@Override
	public Object visit(AstNameExpr nameExpr, Mode mode) {
		if (mode == Mode.BODY) {
			try {
				SemAn.declaredAt.put(nameExpr, symbTable.fnd(nameExpr.name));
			} 
			catch (SymbTable.CannotFndNameException e) {
				throw new Report.Error(nameExpr, "Semantic error: cannot find name " + nameExpr.name);
			}
		}

		return null;
	}

}
